{%- extends "!layout.html" %}

{%- block nav_links %}
<li><a class="nav-link" href="{{ pathto(master_doc) }}">Home</a></li>
<li><a class="nav-link" href="https://docs.aws.amazon.com/cli/latest/userguide/">User Guide</a></li>
<li><a class="nav-link" href="https://forums.aws.amazon.com/forum.jspa?forumID=150">Forum</a></li>
<li><a class="nav-link" href="https://github.com/aws/aws-cli">GitHub</a></li>
{%- endblock %}

{% block ga %}
{% if theme_google_analytics_account %}
<script type="text/javascript" src="https://d36cz9buwru1tt.cloudfront.net/amznUrchin.js"></script>
<!-- SiteCatalyst code version: H.25.2.
  Copyright 1996-2012 Adobe, Inc. All Rights Reserved
  More info available at http://www.omniture.com -->
<script type="text/javascript" src="https://d36cz9buwru1tt.cloudfront.net/js/sitecatalyst/s_code.min.js"></script>
<script type="text/javascript"><!--
s.prop66='AWS CLI'; s.eVar66='D=c66';
s.prop65='User Guide'; s.eVar65='D=c65';
var s_code=s.t();if(s_code)document.write(s_code);
//--></script>
<script type="text/javascript"><!--if(navigator.appVersion.indexOf('MSIE')>=0)document.write(unescape('%3C')+'\!-'+'-')//--></script>
<noscript><img src="https://amazonwebservices.d2.sc.omtrdc.net/b/ss/awsamazondev/1/H.25.2--NS/0" height="1" width="1" border="0" alt=""></noscript><!--/DO NOT REMOVE/--><!-- End SiteCatalyst code version: H.25.2. -->
{% endif %}
{% endblock %}
