# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"). You
 # may not use this file except in compliance with the License. A copy of
 # the License is located at
 #
 # http://aws.amazon.com/apache2.0/
 #
 # or in the "license" file accompanying this file. This file is
 # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
 
from tests import BaseSessionTest
 
 
class TestBedrock(BaseSessionTest):
    def test_invoke_model_with_bidirectional_stream(self):
        """InvokeModelWithBidirectionalStream operation removed due to h2 requirement"""
        bedrock_runtime_client = self.session.create_client('bedrock-runtime', 'us-west-2')
        try:
            bedrock_runtime_client.invoke_model_with_bidirectional_stream()
        except AttributeError:
            pass
        else:
            self.fail(
                'invoke_model_with_bidirectional_stream shouldn\'t be available on the '
                'bedrock-runtime client.'
            )
